from typing import Dict, List, TypeVar

from taskex import Env, TaskRunner

from .group import Group
from .workflow_status import WorkflowStatus

T = TypeVar("T")


class Workflow:
    def __init__(
        self,
        config: Env | None = None,
    ) -> None:
        self._groups: list[Group] = []
        self._execution_orders: Dict[str, List[List[str]]] = {}
        self.runner = TaskRunner(config=config)
        self.status = WorkflowStatus.READY

    def __iter__(self):
        for group in self._groups:
            yield group

    def group(self):
        group = Group(self.runner._snowflake_generator.generate())
        self._groups.append(group)
        return group



import asyncio
import time
import traceback
from collections import defaultdict
from typing import Dict, List

from taskex import Env
from taskex.run import Run
from taskex.util import TimeParser

from .step_type import StepType
from .workflow import Workflow
from .workflow_status import WorkflowStatus


class Orchestral:
    def __init__(
        self,
        config: Env | None = None,
    ) -> None:
        self._workflow = Workflow(config=config)

        self._execution_orders: Dict[str, List[List[str]]] = {}
        self._batch_results: list[Run] = []
        self._results_waiter = asyncio.Event()
        self._run_task: asyncio.Task | None = None
        self._execute_next = False

        self.results: dict[str, dict[int, Run]] = defaultdict(dict)
        self.error: Exception | None = None
        self.trace: str | None = None
        self._start: float | int = 0
        self.elapsed: float = 0

    def __enter__(self):
        self._workflow = Workflow()
        return self._workflow

    def __exit__(self, type, value, traceback):
        pass

    @property
    def status(self):
        if self._workflow:
            return self._workflow.status

    async def __aiter__(self):
        # Before we yield results wait
        if self._results_waiter.is_set() is False:
            await self._results_waiter.wait()

        results = list(self._batch_results)

        yield results

    def create(self):
        self._workflow = Workflow()
        return self._workflow

    async def run(
        self,
        wait: bool = False,
        timeout: str | int | float | None = None,
    ):
        if isinstance(timeout, str):
            timeout = TimeParser(timeout).time

        try:
            self._start = time.monotonic()

            if wait and timeout:
                await asyncio.wait_for(self._run_workflow(), timeout=timeout)

            elif wait:
                await self._run_workflow()

            elif timeout:
                self._run_task = asyncio.create_task(
                    asyncio.wait_for(
                        self._run_workflow(),
                        timeout=timeout,
                    ),
                )

            else:
                self._run_task = asyncio.create_task(
                    self._run_workflow(),
                )

        except asyncio.TimeoutError:
            self.elapsed = time.monotonic() - self._start
            self.trace = traceback.format_exc()
            self.error = Exception(
                f"Err. - Workflow exceeded timeout of {timeout} seconds"
            )

        except Exception as err:
            self.elapsed = time.monotonic() - self._start
            self.trace = traceback.format_exc()
            self.error = err

            self._workflow.status = WorkflowStatus.FAILED

    async def _run_workflow(self):
        self._execute_next = True
        self._workflow.status = WorkflowStatus.READY

        for group in self._workflow:
            self._batch_results.clear()

            group_runs = [
                self._workflow.runner.command(
                    step.call,
                    *step.shell_args,
                    alias=step.alias,
                    env=step.env,
                    cwd=step.cwd,
                    shell=step.shell,
                    timeout=step.timeout,
                    schedule=step.schedule,
                    trigger=step.trigger,
                    repeat=step.repeat,
                    keep=step.keep,
                    max_age=step.max_age,
                    keep_policy=step.keep_policy,
                )
                if step.type == StepType.SHELL
                else self._workflow.runner.run(
                    step.call,
                    alias=step.alias,
                    timeout=step.timeout,
                    schedule=step.schedule,
                    trigger=step.trigger,
                    repeat=step.repeat,
                    keep=step.keep,
                    max_age=step.max_age,
                    keep_policy=step.keep_policy,
                )
                for step in group
            ]

            group_results = await self._workflow.runner.wait_all(
                [run.token for run in group_runs],
            )

            self._batch_results.extend(group_results)

            for res in group_results:
                self.results[res.task_name][res.run_id] = res

            if self._results_waiter.is_set() is False:
                self._results_waiter.set()

            if self._execute_next is False:
                break

        self.elapsed = time.monotonic() - self._start
        self._workflow.status = WorkflowStatus.COMPLETED

    async def stop(self):
        self._execute_next = False
        await self._workflow.runner.stop()

    async def shutdown(self):
        self._execute_next = False
        await self._workflow.runner.shutdown()

        if self._run_task:
            await self._run_task

        if self._results_waiter.is_set() is False:
            self._results_waiter.set()

    def abort(self):
        self._execute_next = False

        try:
            self._run_task.cancel()

        except (
            asyncio.CancelledError,
            asyncio.InvalidStateError,
            asyncio.TimeoutError,
        ):
            pass

        if self._results_waiter.is_set() is False:
            self._results_waiter.set()

        self._workflow.runner.abort()



import inspect
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Literal,
    TypeVar,
)

from .step_type import StepType

T = TypeVar("T")

WorkflowTask = Callable[..., Awaitable[T]] | Callable[..., T]
WorkflowStep = str | WorkflowTask[T]


class Step(Generic[T]):
    def __init__(
        self,
        step: WorkflowStep[T],
        *dependencies: list[WorkflowStep[T]],
        shell_args: tuple[str, ...],
        env: dict[str, Any] | None = None,
        cwd: str | None = None,
        shell: bool = False,
        alias: str | None = None,
        timeout: str | int | float | None = None,
        schedule: str | None = None,
        trigger: Literal["MANUAL", "ON_START"] = "MANUAL",
        repeat: int | Literal["NEVER", "ALWAYS"] = "NEVER",
        keep: int | None = None,
        max_age: str | None = None,
        keep_policy: Literal["COUNT", "AGE", "COUNT_AND_AGE"] = "COUNT",
    ):
        self.call = step
        self.dependencies = [
            dependency
            if isinstance(
                dependency,
                str,
            )
            else dependency.__name__
            for dependency in dependencies
        ]

        self.alias = alias
        self.timeout = timeout
        self.schedule = schedule
        self.trigger = trigger
        self.repeat = repeat
        self.keep = keep
        self.max_age = max_age
        self.keep_policy = keep_policy

        self.env = env
        self.cwd = cwd
        self.shell = shell
        self.shell_args = shell_args

        if isinstance(step, str):
            self.type = StepType.SHELL

        elif (
            inspect.iscoroutine(step)
            or inspect.iscoroutinefunction(step)
            or inspect.isawaitable(step)
        ):
            self.type = StepType.ASYNC

        else:
            self.type = StepType.SYNC



import functools
from typing import Any, Literal, TypeVar

from .step import Step, WorkflowTask

T = TypeVar("T")


class Group:
    def __init__(
        self,
        group_id: int,
    ):
        self.id = group_id
        self._steps: list[Step[T]] = []

    def __iter__(self):
        for step in self._steps:
            yield step

    def add_task(
        self,
        step: WorkflowTask[T],
        *args: tuple[Any, ...],
        alias: str | None = None,
        timeout: str | int | float | None = None,
        schedule: str | None = None,
        trigger: Literal["MANUAL", "ON_START"] = "MANUAL",
        repeat: int | Literal["NEVER", "ALWAYS"] = "NEVER",
        keep: int | None = None,
        max_age: str | None = None,
        keep_policy: Literal["COUNT", "AGE", "COUNT_AND_AGE"] = "COUNT",
        **kwargs: dict[str, Any],
    ):
        self._steps.append(
            Step(
                functools.partial(
                    step,
                    *args,
                    **kwargs,
                ),
                alias=alias,
                timeout=timeout,
                schedule=schedule,
                trigger=trigger,
                repeat=repeat,
                keep=keep,
                max_age=max_age,
                keep_policy=keep_policy,
            )
        )

    def add_command(
        self,
        step: str,
        *args: tuple[str],
        alias: str | None = None,
        env: dict[str, Any] | None = None,
        cwd: str | None = None,
        shell: bool = False,
        timeout: str | int | float | None = None,
        schedule: str | None = None,
        trigger: Literal["MANUAL", "ON_START"] = "MANUAL",
        repeat: int | Literal["NEVER", "ALWAYS"] = "NEVER",
        keep: int | None = None,
        max_age: str | None = None,
        keep_policy: Literal["COUNT", "AGE", "COUNT_AND_AGE"] = "COUNT",
    ):
        self._steps.append(
            Step(
                step,
                shell_args=args,
                env=env,
                cwd=cwd,
                shell=shell,
                alias=alias,
                timeout=timeout,
                schedule=schedule,
                trigger=trigger,
                repeat=repeat,
                keep=keep,
                max_age=max_age,
                keep_policy=keep_policy,
            )
        )



from enum import Enum
from typing import Literal

WorkflowStatusName = Literal["COMPLETED", "FAILED", "RUNNING", "READY"]


class WorkflowStatus(Enum):
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    READY = "READY"
    RUNNING = "RUNNING"
